/* *** Wave file click noise cleaner ***
 * Command Line:
 * cleaner.exe [-{v|vv|vvv}] [wav_file_name [ output_fname [ click_level [ crackle_level [ {p|n}{p|n}... ]]]]]
 *   options:
 *       v   -> verbose log output
 *       vv  -> verbose detailed log output
 *       vvv -> verbose full log output
 *   arguments:
 *       wav_file_name -> 24bit stereo WAVE file name
 *       output_fname  -> cleaned output file
 *       click_level -> 0:normal click noise
 *                1:light click noise
 *                2:havy click noise
 *       crackle_level -> 0:normal crackle noise
 *                1:light crackle noise
 *                2:havy crackle noise
 *       p | n | c -> 1st culumn for L-ch, 2nd culumn for R-ch, pnpn (default)
 *                p: apply positive click noise
 *                n: apply negative click noise
 *                c: apply crackle noise
 *
 * File: cleaner.cpp
 *
 * Created on: 2020/02/14
 * Author    : JR2XZY
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* file pointer and file name pointer */
FILE *fp_log, *fp_wav, *fp_out;

const char *fname_wav = "sample.wav";
const char *fname_log = "sample.log";
const char *fname_out = "sample_out.wav";

/* file buffer pointer */
char *buff = NULL;

/* WAVE file structure definitions */
struct RiffHeader {
	char		riff[4];
	int32_t		size;
	char		type[4];
};

struct FormatChunk{
	char		id[4];
	int32_t		size;
	int16_t		format;
	uint16_t	channels;
	uint32_t	sample_rate;
	uint32_t	byte_per_sec;
	uint16_t	block_align;
	uint16_t	bits_width;
};

struct ExtChunk{
	char		id[4];
	int32_t		size;
	uint8_t		tags[];
};

struct DataChunk{
	char		id[4];
	int32_t		size;
	uint8_t		wave_data[];
};

RiffHeader		*riff;
FormatChunk		*fmt;
ExtChunk		*ext;
DataChunk		*data;
int samples;
int duration;

/* Click Noise Modeling
 *
 *
 *                            P     <--peak level
 *                           + +
 *                           + +
 *                          +  +
 *                         +    +
 *   E.......++++.........+.....+   <--envelope level
 *       ++++    ++++     +     +
 *   ++++            +++++       +
 *                               +                ++
 * 0 ----------------H<==========Z==============+=====>T----
 *                   |            +        +++++       |
 *           head of noise range  +     +++    tail of noise range
 *                                +   ++
 *                                 ++
 * Click noise detection & correction rule:
 * 1. P(noise peak) should be greater than  X(dB) envelope level
 * 2. P(noise peak) shoud be Y(dB) higher than envelope level
 * 3. width of P(noise peak) and Z(zero level) shoud be less than Z(dynamic) samples
 * 4. mirroring at H position until T poinnt
 */

/* constant definition */
#define FLSIZE			100	/* maximum of EnvelopeFilterSize */
#define LEFT_CH			0	/* left channel number */
#define RIGHT_CH		1	/* right channel number */
#define FL_SIZE			0	/* EnvelopeFilterSize */
#define REJECT_PEAK		1	/* RejectPeakRatioForEnvelope (dB) */
#define NOISE_RATIO 	2	/* NoiseRatio (dB) */
#define NOISE_VALUE 	3	/* PeakValue (dB) */
#define MIN_WIDTH		4	/* MinimumNoiseWidth (samples) */
#define NOISE_WIDTH		5	/* NominalNoiseWidth (samples) */
#define MAG_WIDTH		6	/* MagnificationOfNoiseWidth */
#define LARGE_NOISE		7	/* LargePeakLevelForNoiseWidth (dB) */
#define MIRROR_HEAD		8	/* HeadPositionOfMirroring (width) */
#define MIRROR_TAIL		9	/* TailPositionOfMirroring (width) */

/* noise cleaner level parameters table */
int profile1[][10] = {
/* normal click noise */
{	20,		/* EnvelopeFilterSize */
	-6,		/* RejectPeakRatioForEnvelope (dB) */
	10,		/* NoiseRatio (dB) */
	-40,	/* PeakValue (dB) */
	4,		/* MinimumNoiseWidth (samples) */
	30,		/* NominalNoiseWidth (samples) */
	2,		/* MagnificationOfNoiseWidth */
	-10,	/* LargePeakLevelForNoiseWidth (dB) */
	3,		/* HeadPositionOfMirroring (width) */
	6		/* TailPositionOfMirroring (width) */
},
/* light click noise */
{	6,		/* EnvelopeFilterSize */
	-6,		/* RejectPeakRatioForEnvelope (dB) */
	14,		/* NoiseRatio (dB) */
	-20,	/* PeakValue (dB) */
	4,		/* MinimumNoiseWidth (samples) */
	20,		/* NominalNoiseWidth (samples) */
	3,		/* MagnificationOfNoiseWidth */
	-10,	/* LargePeakLevelForNoiseWidth (dB) */
	3,		/* HeadPositionOfMirroring (width) */
	6		/* TailPositionOfMirroring (width) */
},
/* havy click noise */
{	20,		/* EnvelopeFilterSize */
	-6,		/* RejectPeakRatioForEnvelope (dB) */
	10,		/* NoiseRatio (dB) */
	-36,	/* PeakValue (dB) */
	4,		/* MinimumNoiseWidth (samples) */
	20,		/* NominalNoiseWidth (samples) */
	12,		/* MagnificationOfNoiseWidth */
	-10,	/* LargePeakLevelForNoiseWidth (dB) */
	3,		/* HeadPositionOfMirroring (width) */
	6		/* TailPositionOfMirroring (width) */
}
};

double reject_peak;		/* RejectPeakRatio */
double noise_ratio;		/* NoiseRatio */
double noise_value;		/* PeakValue */
int noise_db;			/* PeakValue (dB) */
int min_width;			/* MinimumNoiseWidth (samples) */
int noise_width;		/* NominalNoiseWidth (samples) */
int magnification;		/* MagnificationOfNoiseWidth */
int large_peak;			/* LargePeakLevelForNoiseWidth */
int mirror_head;		/* HeadPositionOfMirroring (width) */
int mirror_tail;		/* TailPositionOfMirroring (width) */

/* Cracle Noise Modeling
 *
 *
 *                            +
 *                           + +
 *                           + +
 *                          +  +
 *                         +    +
 *           ++++         +  P   +      <--peak level
 *       ++++    ++++     + *    +
 *   ++++            +++++ *      +   ++  +
 *       *   *            *         + * ++ +++++++++
 * 0 ****-***-***-***-****----*-----**-*-***-------------
 *               *   * |       *   *   +*
 *          H<============Z==================>T
 *          |                                 |
 *           head of noise rang+e**           tail of noise range
 *
 *
 * Crackle noise detection & correction rule:
 * 1. caluculate differential
 * 2. apply high pass filter
 * 3. detect contiguous peaks for noise range
 * 4. determine maximum peak in the range should be greater than  X(dB) envelope level
 * 5. interpolating from H position to T poinnt to mask crackle noise
 */

/* constant definition */
#define CUTOFF_FREQ		0	/* CutOffFrequency (Hz) */
#define Q_FACTOR	 	1	/* QualityFactor (0.001 unit) */
#define DIFF_GAIN		2	/* DifferentialGain (dB) */
#define PEAK_LEVEL		3	/* PeakLevel (dB) */
#define PEAK_DURATION	4	/* Minimum Peak duration (samples) */
#define MASK_HEAD		5	/* HeadPositionOfMasking (width) */
#define MASK_TAIL		6	/* TailPositionOfMasking (width) */

/* noise cleaner level parameters table */
int profile2[][8] = {
/* normal crackle noise */
{	20000,	/* CutOffFrequency (Hz) */
	1000,	/* QualityFactor (0.001 unit) */
	-6,		/* DifferentialGain (dB) */
	-62,	/* PeakLevel (dB) */
	10,		/* Minimum Peak duration (samples) */
	3,		/* HeadPositionOfMasking (width) */
	2		/* TailPositionOfMasking (width) */
},
/* light click noise */
{	18000,	/* CutOffFrequency (Hz) */
	1000,	/* QualityFactor (0.001 unit) */
	-6,		/* DifferentialGain (dB) */
	-70,	/* PeakLevel (dB) */
	8,		/* Minimum Peak duration (samples) */
	2,		/* HeadPositionOfMasking (width) */
	2		/* TailPositionOfMasking (width) */
},
/* havy click noise */
{	20000,	/* CutOffFrequency (Hz) */
	1000,	/* QualityFactor (0.001 unit) */
	-6,		/* DifferentialGain (dB) */
	-48,	/* PeakLevel (dB) */
	10,		/* Minimum Peak duration (samples) */
	3,		/* HeadPositionOfMasking (width) */
	2		/* TailPositionOfMasking (width) */
}
};

double cut_freq;	/* CutOffFrequency (Hz) */
double q_factor;	/* QualityFactor (0.001 unit) */
double diff_gain;	/* DifferentialGain (dB) */
int peak_level;		/* PeakLevel (dB) */
int peak_duration;	/* Minimum Peak duration (samples) */
int mask_head;		/* HeadPositionOfMasking (width) */
int mask_tail;		/* TailPositionOfMasking (width) */

int verbose = 0;			/* verbose log output flag */
int click_level = 0;		/* click noise cleaner level */
int crackle_level = 0;		/* crackle noise cleaner level */
int fl_size;				/* EnvelopeFilterSize */
const char *sequence;		/* optional sequence */
int level_0db = 0x7fffff;	/* 24bits PCM */

int get_sample(int ch, int i) {
	if(i < 0 || i >= samples) return 0;
	int s;
	int k = i * fmt->block_align + ch * 3;
	s = (data->wave_data[k + 2] << 16) + (data->wave_data[k + 1] << 8) + data->wave_data[k];
	s |= (s & 0x800000)?0xff000000:0;
	return s;
}

int set_sample(int ch, int i, int v) {
	if(i < 0) i = 0;
	if(i >= samples) i = samples - 1;
	int k = i * fmt->block_align + ch * 3;
	data->wave_data[k] = v & 0xff;
	data->wave_data[k + 1] = (v >> 8) & 0xff;
	data->wave_data[k + 2] = v >> 16;
	return v;
}

/* differential filtered sample */
int last_sample;

int differential(int ch, int i) {
	int sample = get_sample(ch, i);
	int diff = (int)(diff_gain * (sample - last_sample));
	last_sample = sample;
	return diff;
}

/* 12dB/Oct High Pass Filter */
double a0, a1, a2;		/* filter coefficients */
double b0, b1, b2;		/* filter coefficients */
double x1, x2, f1, f2;	/* work area */

void setup_hpf(int srate, double freq, double q) {
	double o = 2.0 * M_PI *  freq / srate;
	double a = sin(o) / (2.0 * q);
	a0 =   1.0 + a;
	a1 =  -2.0 * cos(o);
	a2 =   1.0 - a;
	b0 =  (1.0 + cos(o)) / 2.0;
	b1 = -(1.0 + cos(o));
	b2 =  (1.0 + cos(o)) / 2.0;
	x1 = x2 = f1 = f2 = 0.0;
	last_sample = 0;
}

int high_pass(int x) {
 	double f = b0 / a0 * x + b1 / a0 * x1 + b2 / a0 * x2 - a1 / a0 * f1 - a2 / a0 * f2;
	x2 = x1;
	x1 = x;
	f2 = f1;
	f1 = f;
	return (int)f;
}

void interpolate(int ch, int p, int head, int tail, int width) {
	int delta = get_sample(ch, p + tail) - get_sample(ch, p + head);
	for(int i = head; i < tail; i++) {
		double lpf = 0.0;
		for(int j = - width; j <  width; j++) {
			if((i + j> head) && (i + j > tail)) {
				lpf +=  get_sample(ch, p + head) + delta * (i + j - head) / (tail - head);
			} else {
				lpf +=  get_sample(ch, p + i + j);
			}
		}
		set_sample(ch, p + i, (int)(lpf / 2.0 / width));
	}
}

int crackle_noise(int ch) {
	int sample, last, peak, p, p1, p2;
    int width;
    int n;
    int noises = 0;
    /* positive peak detection */
	peak = 0;
	p = p1 = p2 = 0;
	n = 0;
    for(int i = 0; i < samples; i++) {
		sample = high_pass(differential(ch, i));
		double time = int(1000000.0 * i / fmt->sample_rate) / 1000.0;
		if(verbose > 2) {
			set_sample(ch, i, sample);
			continue;
		}
		if(verbose > 1) printf("%d, %.3f, %d, %d, %d, %d\n", i, time, sample, peak, p, n);
		if(sample > 0) { /* positive peak detection */
			if(last < 0) {
				if(i - p > peak_duration) {
					if(n > 0) n = - n; else p = p1 = p2 = i;
				} else {
					p2 = i;
					n++;
				}
			}
			if(sample > peak) {
				peak = sample;
				p = i;
			}
		}
		if(sample < 0) { /* once negative sample */
			if((n < 0) && (p1 != p2)) { /* once negative sample */
				width = p2 - p1; /* peak to zero width */
				/* check envelope peak ratio, peak value and peak to zero width in order to detect click noise */
				if((peak >  peak_level) && (width < peak_duration)) {
					double time = int(1000000.0 * p1 / fmt->sample_rate) / 1000.0;
					if(verbose) printf("%d, %.3f, %d, %d\n", p1, time, peak, width);
					int head = - width * mask_head;
					int tail = width * mask_tail;
					interpolate(ch, p1, head, tail, width);
#if 0
					int head = - width * mask_head;
					int tail = width * mask_tail;
					int delta = get_sample(ch, p1 + tail) - get_sample(ch, p1 + head);
					for(int k = head; k < tail; k++) {
						set_sample(ch, p1 + k, get_sample(ch, p1 + head) + delta * (k - head) / (tail - head));
					}
#endif
					i = p1 + tail;
					noises++;
					p = i;
				}
				peak = 0;
				n = 0;
			}
		}
		last = sample;
	}
	return noises;
}

int click_negative(int ch) {
	int sample, peak, p;
    int peaks[FLSIZE];
    int envelope;
    int width;
    int k, n;
    int noises = 0;
    /* negative peak detection */
    envelope = 0;
	peak = 0;
	p = -1;
	k = n = 0;
    for(int i = 0; i < samples; i++) {
		sample = get_sample(ch, i);
		if(verbose > 1) printf("%d, %d\n", i, sample);
		if(sample < 0) { /* negative peak detection */
			if(peak > sample) {
				peak = sample;
				p = i;
			}
		}
		if((p >= 0) && (sample >= 0)) { /* once positive sample */
			if(n) { /* calcurate envelope level */
				double sum = 0.0;
				for(int j = 0; j < n; j++) {
					sum += peaks[j];
				}
				envelope = sum / n;
			}
			width = i - p; /* peak to zero width */
			/* check envelope peak ratio, peak value and peak to zero width in order to detect click noise */
			double coeff = (peak - envelope + noise_value) / (-large_peak + noise_value);
			coeff = (coeff < 0)?1:(coeff > 1)?magnification:coeff * (magnification - 1) + 1;
			if((peak < envelope * noise_ratio) && ((peak - envelope) < -noise_value) &&
					(width < noise_width * coeff) && (width > min_width)) {
				double time = int(1000000.0 * p / fmt->sample_rate) / 1000.0;
				if(verbose) printf("%d, %.3f, %d, %d, %d, %d\n", p, time, peak, envelope, width, (int)(noise_width * coeff));
#if 0
				int head = - width * mirror_head;
				int tail = width * mirror_tail;
				interpolate(ch, p, head, tail, width);
#else
				int head = - width * mirror_head;
				int tail = width * mirror_tail;
				for(int k = head; k < tail; k++) {
					set_sample(ch, p + k, get_sample(ch, p + head - (k - head)));
				}
#endif
				noises++;
			}
			if(peak < envelope * reject_peak) { /* accumlate peak */
			peaks[k] = peak;
			k = (k < fl_size - 1)?k+1:0;
			n = (n < fl_size)?n+1:n;
			}
			peak = 0;
			p = -1;
		}
	}
	return noises;
}

int click_positive(int ch) {
	int sample, peak, p;
    int peaks[FLSIZE];
    int envelope;
    int width;
    int k, n;
    int noises = 0;
    /* positive peak detection */
    envelope = 0;
	peak = 0;
	p = -1;
	k = n = 0;
    for(int i = 0; i < samples; i++) {
		sample = get_sample(ch, i);
		if(verbose > 1) printf("%d, %d\n", i, sample);
		if(sample > 0) { /* positive peak detection */
			if(peak < sample) {
				peak = sample;
				p = i;
			}
		}
		if((p >= 0) && (sample < 0)) { /* once negative sample */
			if(n) { /* calcurate envelope level */
				double sum = 0.0;
				for(int j = 0; j < n; j++) {
					sum += peaks[j];
				}
				envelope = sum / n;
			}
			width = i - p; /* peak to zero width */
			/* check envelope peak ratio, peak value and peak to zero width in order to detect click noise */
			double coeff = (peak - envelope - noise_value) / (large_peak - noise_value);
			coeff = (coeff < 0)?1:(coeff > 1)?magnification:coeff * (magnification - 1) + 1;
			if((peak > envelope * noise_ratio) && ((peak - envelope) > noise_value) &&
					(width < noise_width * coeff) && (width > min_width)) {
				double time = int(1000000.0 * p / fmt->sample_rate) / 1000.0;
				if(verbose) printf("%d, %.3f, %d, %d, %d, %d\n", p, time, peak, envelope, width, (int)(noise_width * coeff));
#if 0
				int head = - width * mirror_head;
				int tail = width * mirror_tail;
				interpolate(ch, p, head, tail, width);
#else
				int head = - width * mirror_head;
				int tail = width * mirror_tail;
				for(int k = head; k < tail; k++) {
					set_sample(ch, p + k, get_sample(ch, p + head - (k - head)));
				}
#endif
				noises++;
			}
			if(peak > envelope * reject_peak) { /* accumlate peak */
			peaks[k] = peak;
			k = (k < fl_size - 1)?k+1:0;
			n = (n < fl_size)?n+1:n;
			}
			peak = 0;
			p = -1;
		}
	}
	return noises;
}

#define LEVEL_RANGE 60
#define WIDTH_RANGE 500

int statistics(int ch) {
	int peak_stat[WIDTH_RANGE][LEVEL_RANGE];
	int sample, peak, p;
    int peaks = 0;
    for(int i =0; i < WIDTH_RANGE;  i++) {
    	for(int j =0; j < LEVEL_RANGE; j++) peak_stat[i][j] = 0;
    }
    /* positive peak detection */
	peak = 0;
	p = -1;
    for(int i = 0; i < samples; i++) {
		sample = get_sample(ch, i);
		if(sample > 0) { /* positive peak detection */
			if(peak < sample) {
				peak = sample;
				p = i;
				peaks++;
			}
		}
		if((p >= 0) && (sample < 0)) { /* once negative sample */
			int width = i - p - 1; /* peak to zero width */
			if(width >= WIDTH_RANGE) width = WIDTH_RANGE -1;
			int db = -20.0 * log10(peak / (double)level_0db);
			if(db < 0) db = 0;
			if(db >= LEVEL_RANGE) db = LEVEL_RANGE -1;
			peak_stat[width][db]++;
			peak = 0;
			p = -1;
		}
	}
    for(int i =0; i < WIDTH_RANGE;  i++) {
    	printf("%d points ->", i + 1);
    	for(int j =0; j < LEVEL_RANGE; j++) printf(" %d", peak_stat[i][j]);
    	printf("\n");
    }
	return peaks;
}

int main(int argc, char *argv[]) {

	printf("*** Wave file click noise cleaner ***\n");

	/* Command line argments:
	 * cleaner.exe [-v] [wav_file_name [ output_fname [level]]]
	 */
	int arg = 1;
	printf("option: ");
	for(int i = 1; i < argc; i++) printf("%s ", argv[i]);
	printf("\n");
	if(argc == 1) {
		const char *help =
		"Command Line:\n"
		" cleaner.exe [-{v|vv|vvv}] [wav_file_name [ output_fname [ click_level [ crackle_level [ {p|n}{p|n}... ]]]]]\n"
		" options:\n"
		"   v   -> verbose log output\n"
		"   vv  -> verbose detailed log output\n"
		"   vvv -> verbose full log output\n"
		" arguments:\n"
		"   wav_file_name -> 24bit stereo WAVE file name\n"
		"   output_fname  -> cleaned output file\n"
		"   click_level -> 0:normal click noise\n"
		"            1:small click noise\n"
		"            2:big click noise\n"
		"   crackle_level -> 0:normal crackle noise\n"
		"            1:small crackle noise\n"
		"            2:big crackle noise\n"
		"   p | n | c -> 1st culumn for L-ch, 2nd culumn for R-ch, pnpn (default)"
		"            p: apply positive click noise\n"
		"            n: apply negative click noise\n"
		"            c: apply crackle noise\n";
		printf(help);
		return -1;
	}
	if(argc > 1) {
		if(argv[1][0] == '-') {
		  if(!strncmp(argv[1], "-v",   3)) verbose = 1;
		  if(!strncmp(argv[1], "-vv",  4)) verbose = 2;
		  if(!strncmp(argv[1], "-vvv", 5)) verbose = 3;
		  arg++;
		  argc--;
		}
	}
	if(argc > 1) fname_wav = argv[arg++];
	if(argc > 2) fname_out = argv[arg++];
	if(argc > 3) {
		if(!strncmp(argv[arg], "0", 2)) click_level = 0;	/* normal click noise */
		if(!strncmp(argv[arg], "1", 2)) click_level = 1;	/* light  click noise */
		if(!strncmp(argv[arg], "2", 2)) click_level = 2;	/* havy   click noise */
		arg++;
	}
	if(argc > 4) {
		if(!strncmp(argv[arg], "0", 2)) crackle_level = 0;	/* normal crackle noise */
		if(!strncmp(argv[arg], "1", 2)) crackle_level = 1;	/* light  crackle noise */
		if(!strncmp(argv[arg], "2", 2)) crackle_level = 2;	/* havy   crackle noise */
		arg++;
	}
	sequence = "pnnp";
	if(argc > 5) {
		sequence = argv[arg];
	}


	/* Reading WAV file and check format parameters */
	fpos_t f_length;
    uint32_t len;

  	fp_wav = fopen(fname_wav, "rb");
	if(fp_wav == NULL) {
	  printf( "%s file not opened.\n", fname_wav);
	  return -1;
	}
	if(!fseek(fp_wav, 0L, SEEK_END)) {
		if(!fgetpos(fp_wav, &f_length)) {
        	if(f_length == 0) {
        		printf( "%s zero length file.\n", fname_wav);
        		return -1;
        	}
       	}
    } else {
        printf( "%s fgetpos error.¥n", fname_wav);
        return -1;
    }
	buff = (char*)malloc(f_length);
 	fseek(fp_wav, 0L, 0L);
 	len = fread(buff, sizeof(char), f_length, fp_wav);
    printf( "Input wave file: %s -> %d bytes read.\n", fname_wav, len );
    fclose(fp_wav);

    riff = (RiffHeader*)buff;
    fmt = (FormatChunk*)(buff + sizeof(RiffHeader));
    ext = (ExtChunk*)((char*)fmt + sizeof(FormatChunk));
    if(ext->id[0] != 'b' || ext->id[1] != 'e' || ext->id[2] != 'x' || ext->id[3] != 't') {
    	data = (DataChunk*)(ext);
    	ext = NULL;
	} else {
		data = (DataChunk*)((char*)ext + ext->size + 8);
	}
	samples = data->size / fmt->block_align;
    duration = 1000000.0 * data->size / fmt->block_align / fmt->sample_rate;

	printf( "\nHeader Information:\n");
    printf( "RIFF type    = %c%c%c%c\n", riff->type[0], riff->type[1], riff->type[2], riff->type[3]);
    printf( "format id    = %c%c%c%c\n", fmt->id[0], fmt->id[1], fmt->id[2], fmt->id[3]);
    printf( "format       = 0x%04X\n", fmt->format);
    printf( "channels     = %d\n", fmt->channels);
    printf( "sample rate  = %d\n", fmt->sample_rate);
    printf( "byte per sec = %d\n", fmt->byte_per_sec);
    printf( "block align  = %d\n", fmt->block_align);
    printf( "bits width   = %d\n", fmt->bits_width);
    printf( "duration     = %ds %3dms %3dus\n", duration / 1000000, (duration % 1000000) / 1000, duration % 1000);
    if((fmt->format != 0x0001) || (fmt->channels != 2) || (fmt->bits_width != 24)) {
    	printf("Format Error!!\n");
		return -1;
	}

	if(verbose > 2) {
		printf("\nL-ch statistics\n");
		statistics(LEFT_CH);
		printf("\nR-ch statistics\n");
		statistics(RIGHT_CH);
	}

    /* click niose cleaner parameters */
    fl_size       = profile1[click_level][FL_SIZE];
	reject_peak   = pow(10.0, profile1[click_level][REJECT_PEAK] / 20.0);
	noise_ratio   = pow(10.0, profile1[click_level][NOISE_RATIO] / 20.0);
	noise_value   = pow(2.0, fmt->bits_width - 1) * pow(10.0, profile1[click_level][NOISE_VALUE] / 20.0);
	noise_db      = profile1[click_level][NOISE_VALUE];
	min_width     = profile1[click_level][MIN_WIDTH];
	noise_width   = profile1[click_level][NOISE_WIDTH];
	magnification = profile1[click_level][MAG_WIDTH];
	large_peak    = pow(2.0, fmt->bits_width - 1) * pow(10.0, profile1[click_level][LARGE_NOISE] / 20.0);
	mirror_head   = profile1[click_level][MIRROR_HEAD];
	mirror_tail   = profile1[click_level][MIRROR_TAIL];

	/* crackle noise cleaner parameter */
	cut_freq      = profile2[crackle_level][CUTOFF_FREQ];
	q_factor      = profile2[crackle_level][Q_FACTOR] / 1000.0;
	diff_gain     = pow(10.0, profile2[crackle_level][DIFF_GAIN] / 20.0);
	peak_level    = pow(2.0, fmt->bits_width - 1) * pow(10.0, profile2[crackle_level][PEAK_LEVEL] / 20.0);
	peak_duration = profile2[crackle_level][PEAK_DURATION];
	mask_head     = profile2[crackle_level][MASK_HEAD];
	mask_tail     = profile2[crackle_level][MASK_TAIL];

	printf("\nCleaning ...\n");
    for(int i = 0; i < (int)strlen(sequence); i++) {
    	char c = sequence[i];
    	if(c == 'p') {
    		if(i & 1) printf("R-ch %d positive click noises cleaned.\n", click_positive(RIGHT_CH));
    		  else    printf("L-ch %d positive click noises cleaned.\n", click_positive(LEFT_CH));
    	}
    	if(c == 'n') {
    		if(i & 1) printf("R-ch %d negative click noises cleaned.\n", click_negative(RIGHT_CH));
    		  else    printf("L-ch %d negative click noises cleaned.\n", click_positive(LEFT_CH));
    	}
    	if(c == 'c') {
    		setup_hpf(fmt->sample_rate, cut_freq, q_factor);
    		if(i & 1) printf("R-ch %d crackle noises cleaned.\n", crackle_noise(RIGHT_CH));
    		  else    printf("L-ch %d crackle noises cleaned.\n", crackle_noise(LEFT_CH));
    	}
	}

    /* writing cleaned data */
  	fp_out = fopen(fname_out, "wb");
	if(fp_out == NULL) {
	  printf( "%s file not opened.\n", fname_out);
	  return -1;
	}
 	len = fwrite(buff, sizeof(char), f_length, fp_out);
    printf( "\nOutput wave file: %s -> %d bytes written.\n", fname_out, len );
    fclose(fp_out);

    free(buff);
}



